from acrylic.Schemes import scheme
from acrylic.Defaults import SCHEMAS
from acrylic.Validators import validate
from acrylic.Converters import rgb_to, rgb_from


class ColorMeta(type):
    def __new__(cls, clsname, bases, attrs):

        # generates readonly, lazy evaluation properties
        def property_factory(attr):

            def getter(instance):
                value = getattr(instance, f'_{attr}')

                if value:
                    # value for `attr` is already present, send cached value
                    return value

                if value is None and attr == 'rgb':
                    # rgb for this color has not been calculated yet
                    # convert color's default to rgb and cache
                    default = getattr(instance, instance._default)
                    converted = rgb_from[instance._default](default)
                    instance._rgb = converted
                    return converted

                # `attr` for this color has not been calculated yet
                # convert to `attr` from rgb and cache
                converted = rgb_to[attr](instance.rgb)
                setattr(instance, f'_{attr}', converted)
                return converted

            def setter(instance, *args, **kwargs):
                raise AttributeError(f'{attr!r} is a readonly attribute')

            doc = f'(readonly) return {attr} component of the color'
            return property(getter, setter, doc=doc)

        for colorspace in SCHEMAS:
            # add properties for each colorspace using factory (Color.rgb)
            attrs[colorspace] = property_factory(colorspace)

            # add private attributes for each colorspace property (Color._rgb)
            attrs[f'_{colorspace}'] = None

        return super(ColorMeta, cls).__new__(cls, clsname, bases, attrs)


class Color(metaclass=ColorMeta):
    '''
    class to represent colors

    Supported color formats: rgb, hsl, hsv, hex, name, ryb
    check `help(acrylic.Color)` for creating a new Color instance

    '''

    def __init__(self, **kwargs):
        '''
        Create an instance of `Color`

        Supported formats:
        rgb, hsl, hsv, ryb, hex, name

        Examples:
            >>> from acrylic import Color, RANDOM

            Create a color using rgb
            >>> Color(rgb=[83, 237, 229])
            Color(rgb=(83, 237, 229))

            Create a color using hsl with a range of saturation values
            >>> Color(hsl=[176.5, (20, 70), 62])
            Color(hsl=(176.5, 51.58, 62))

            Note: Since saturation is picked from the range (20, 70),
            it can differ from 51.58

            Create a color with random hue
            >>> Color(hsv=[RANDOM, 75, 98])
            Color(hsv=(167.23, 75, 98))

            Note: Since hue is random it can differ from 167.23

            Create a color using hex
            >>> Color(hex='#80ffd4')
            Color(hex='#80ffd4')

            Create a color using a valid CSS3 name
            >>> Color(name='aquamarine')
            Color(name='aquamarine')

        Args:
            rgb (optional):
                Iterable with 3 values
                0 <= {r, g, b} <= 255
            ryb (optional):
                Iterable with 3 values
                0 <= {r, y, b} <= 255
            hsv (optional):
                Iterable with 3 values
                0.0 <= {h} <= 360.0
                0.0 <= {s, v} <= 100.0
            hsl (optional):
                Iterable with 3 values
                0.0 <= {h} <= 360.0
                0.0 <= {s, l} <= 100.0
            hex (optional):
                String representing 6-digit hex number
            name (optional):
                String representing a valid CSS3 color name

        Note:
            Each value of the Iterable can be:
                - a single value
                - a list of 2 values, a random value will be picked from
                  between these 2 values
                - `Color.RANDOM` or -1, a random value from that component's
                  range will be picked

        Raises:
            TypeError: if datatypes dont match
            ValueError: values are not within valid ranges
        '''

        if len(kwargs) > 1:
            raise TypeError('Color() got multiple keyword arguments')

        if kwargs:
            colorspace, value = list(kwargs.items())[0]
        else:
            colorspace, value = 'rgb', (0, 0, 0)

        if colorspace not in SCHEMAS.keys():
            msg = f'Color() got an unexpected keyword argument {colorspace!r}'
            raise TypeError(msg)

        # set the private attribute for this colorspace's property
        # ex: set _rgb for Color().rgb
        setattr(self, f'_{colorspace}', validate(value, colorspace))

        # set default format to use for repr() and str() and lazy evaluation
        self._default = colorspace

    def __repr__(self):
        template = f'{type(self).__name__}({self._default}={{}})'

        if SCHEMAS[self._default].validation_type == 'values':
            values = tuple(getattr(self, self._default))
            return template.format(repr(values))

        if SCHEMAS[self._default].validation_type == 'string':
            return template.format(repr(getattr(self, self._default)))

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.rgb == other.rgb
        return NotImplemented

    def __hash__(self):
        return hash(self.rgb)

    # - - - - - - - - - -

    def _in_ryb(self):
        new_hue = Color(ryb=self.rgb).hsl.h
        return Color(hsl=(new_hue, self.hsl.s, self.hsl.l))

    scheme = scheme



import colorsys

from acrylic.Defaults import SCHEMAS, PRECISION
from acrylic.color_names import color_names, color_names_reverse


def rgb_to_hsl(rgb):
    rgb_format, hsl_format = SCHEMAS['rgb'].format, SCHEMAS['hsl'].format

    norm = [x / m[1] for x, m in zip(rgb, rgb_format)]
    h, l, s = colorsys.rgb_to_hls(*norm)

    hsl = [round(x * m[1], PRECISION) for x, m in zip([h, s, l], hsl_format)]
    return SCHEMAS['hsl'].output_type(*hsl)


def rgb_to_hsv(rgb):
    rgb_format, hsv_format = SCHEMAS['rgb'].format, SCHEMAS['hsv'].format

    norm = [x / m[1] for x, m in zip(rgb, rgb_format)]
    h, s, v = colorsys.rgb_to_hsv(*norm)

    hsv = [round(x * m[1], PRECISION) for x, m in zip([h, s, v], hsv_format)]
    return SCHEMAS['hsv'].output_type(*hsv)


def rgb_to_hex(rgb):
    return SCHEMAS['hex'].output_type(''.join(f'{x:02X}' for x in rgb))


def rgb_to_name(rgb):
    return color_names_reverse[rgb] if rgb in color_names_reverse else '-'


def rgb_to_ryb(rgb):
    # Source:
    # https://www.jstage.jst.go.jp/article/tievciieej/5/2/5_110/_pdf/-char/en

    rgb_format, ryb_format = SCHEMAS['rgb'].format, SCHEMAS['ryb'].format
    rgb_r, rgb_g, rgb_b = [x / m[1] for x, m in zip(rgb, rgb_format)]

    white = min(rgb_r, rgb_g, rgb_b)
    black = min(1 - rgb_r, 1 - rgb_g, 1 - rgb_b)
    rgb_r, rgb_g, rgb_b = [x - white for x in (rgb_r, rgb_g, rgb_b)]

    yellow = min(rgb_r, rgb_g)
    ryb_r = rgb_r - yellow
    ryb_y = (yellow + rgb_g) / 2
    ryb_b = (rgb_b + rgb_g - yellow) / 2

    norm = 0
    if max(rgb_r, rgb_g, rgb_b) != 0:
        norm = max(ryb_r, ryb_y, ryb_b) / max(rgb_r, rgb_g, rgb_b)

    ryb = [x / norm if norm > 0 else x for x in (ryb_r, ryb_y, ryb_b)]
    ryb = [x + black for x in ryb]

    result = [round(x * m[1]) for x, m in zip(ryb, ryb_format)]
    return SCHEMAS['ryb'].output_type(*result)


# - - - - - - - - - - - - - -


def hsl_to_rgb(hsl):
    rgb_format, hsl_format = SCHEMAS['rgb'].format, SCHEMAS['hsl'].format

    (h, s, l) = [x / m[1] for x, m in zip(hsl, hsl_format)]
    rgb = colorsys.hls_to_rgb(h, l, s)

    rgb = [round(x * m[1]) for x, m in zip(rgb, rgb_format)]
    return SCHEMAS['rgb'].output_type(*rgb)


def hsv_to_rgb(hsv):
    rgb_format, hsv_format = SCHEMAS['rgb'].format, SCHEMAS['hsv'].format

    (h, s, v) = [x / m[1] for x, m in zip(hsv, hsv_format)]
    rgb = colorsys.hsv_to_rgb(h, s, v)

    rgb = [round(x * m[1]) for x, m in zip(rgb, rgb_format)]
    return SCHEMAS['rgb'].output_type(*rgb)


def hex_to_rgb(hex_str):
    rgb = [int(hex_str[x:x + 2], 16) for x in range(1, 7, 2)]
    return SCHEMAS['rgb'].output_type(*rgb)


def name_to_rgb(name):
    if name == '-':
        return SCHEMAS['rgb'].output_type(0, 0, 0)
    return SCHEMAS['rgb'].output_type(*color_names[name])


def ryb_to_rgb(ryb):
    # Source:
    # https://www.jstage.jst.go.jp/article/tievciieej/5/2/5_110/_pdf/-char/en

    rgb_format, ryb_format = SCHEMAS['rgb'].format, SCHEMAS['ryb'].format
    ryb_r, ryb_y, ryb_b = [x / m[1] for x, m in zip(ryb, ryb_format)]

    black = min(ryb_r, ryb_y, ryb_b)
    white = min(1 - ryb_r, 1 - ryb_y, 1 - ryb_b)
    (ryb_r, ryb_y, ryb_b) = (x - black for x in (ryb_r, ryb_y, ryb_b))

    green = min(ryb_y, ryb_b)
    rgb_r = ryb_r + ryb_y - green
    rgb_g = ryb_y + green
    rgb_b = 2 * (ryb_b - green)

    norm = 0
    if max(ryb_r, ryb_y, ryb_b) != 0:
        norm = max(rgb_r, rgb_g, rgb_b) / max(ryb_r, ryb_y, ryb_b)

    rgb = [x / norm if norm > 0 else x for x in (rgb_r, rgb_g, rgb_b)]
    rgb = [x + white for x in rgb]

    result = [round(x * m[1]) for x, m in zip(rgb, rgb_format)]
    return SCHEMAS['rgb'].output_type(*result)


# - - - - - - - - - - - - - -


rgb_to = {
    'hsl': rgb_to_hsl,
    'hsv': rgb_to_hsv,
    'hex': rgb_to_hex,
    'name': rgb_to_name,
    'ryb': rgb_to_ryb,
}


rgb_from = {
    'hsl': hsl_to_rgb,
    'hsv': hsv_to_rgb,
    'hex': hex_to_rgb,
    'name': name_to_rgb,
    'ryb': ryb_to_rgb,
}



'''
acrylic is a simple and intutive library to work with color in python. 
It can be used to handle color data, easily convert between different 
color spaces and work with colorschemes. acrylic currently supports the 
following color formats:
rgb, hsl, hsv, hex, name, ryb

Try this for more details on how to use acrylic:
>>> from acrylic import Color
>>> help(Color)
'''


__version__ = '0.3.0'


from .Color import Color
from .Defaults import RANDOM
from . import Schemes



from random import uniform, randint

from acrylic.Defaults import RANDOM, SCHEMAS
from acrylic.Validators import in_range, check_datatype


ANALOGOUS = 1
COMPLEMENTARY = 2
TRIADIC = TRIANGLE = 3
TETRADIC = SQUARE = 4
MONOCHROMATIC = 5
SHADES = 6
SPLIT_COMPLEMENTARY = 7
ACCENTED_ANALOGOUS = ANALOGOUS_COMPLEMENTARY = 8
RECTANGLE = 9
NEAR_COMPLEMENTARY = 10
COMPLEMENTARY_TRIADIC = 11
MODIFIED_TRIADIC = 12


def scheme(self, name, in_rgb=False, fuzzy=0):
    '''
    Returns a list of `Color` according to the given color scheme

    Args:
        name(int):
            a variable from `acrylic.Schemes`
            example: Schemes.COMPLEMENTARY
        in_rgb(bool, optional):
            False: use the RYB color wheel for calculations
            True: use the RGB color wheel for calculations
            Default: False
        fuzzy(int, optional):
            adds a random value between -fuzzy and +fuzzy to
            the generated hue. Can be set to `RANDOM` to use
            the recommended value.
            Default: 0

    Returns:
        list of Color
    '''
    Color = type(self)

    colors, (a, b) = list(), SCHEMAS['hsl'].format.h
    if fuzzy == RANDOM:
        fuzzy = uniform(b / 20, b / 5) if randint(0, 5) else 0
    else:
        fuzzy = check_datatype(float, fuzzy, 'fuzzy')
        fuzzy = in_range(fuzzy, a, b, 'fuzzy')

    deltas = {
        ANALOGOUS: [b / 24, -(b / 24), b / 12, -(b / 12)],
        COMPLEMENTARY: [b / 2],
        TRIADIC: [b / 3, -(b / 3)],
        SQUARE: [b / 4, -(b / 4), b / 2],
        SPLIT_COMPLEMENTARY: [(b / 2) - (b / 12), (b / 2) + (b / 12)],
        ACCENTED_ANALOGOUS: [b / 12, -(b / 12), b / 2],
        RECTANGLE: [b / 2, (b / 2) - (b / 8), b / 8],
        NEAR_COMPLEMENTARY: [(b / 2) - (b / 12)],
        COMPLEMENTARY_TRIADIC: [b / 2, b / 4],
        MODIFIED_TRIADIC: [b / 12, b / 6],
    }

    if name == MONOCHROMATIC:
        m = (SCHEMAS['hsv'].format.s[1] / (SCHEMAS['hsv'].format.s[1] / 30))
        if self.hsv.s > m:
            s = [self.hsv.s, self.hsv.s - m, self.hsv.s, self.hsv.s - m]
        else:
            s = [self.hsv.s, self.hsv.s + m, self.hsv.s, self.hsv.s + m]
        v = [SCHEMAS['hsv'].format.v[1] / 5, SCHEMAS['hsv'].format.v[1]]
        for i in range(4):
            colors.append(Color(hsv=(self.hsl.h, s[i], v)))
        return colors

    elif name == SHADES:
        v = [SCHEMAS['hsv'].format.v[1] / 5, SCHEMAS['hsv'].format.v[1]]
        for i in range(4):
            colors.append(Color(hsv=(self.hsl.h, self.hsv.s, v)))
        return colors

    elif name not in deltas:
        raise ValueError(f'{name!r} is not a valid color scheme')

    for angle in deltas[name]:
        h = (self.hsl.h + angle + uniform(-fuzzy, +fuzzy)) % b
        if not in_rgb:
            colors.append(Color(hsv=(h, self.hsv.s, self.hsv.v))._in_ryb())
        else:
            colors.append(Color(hsv=(h, self.hsv.s, self.hsv.v)))
    return colors
